#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <math.h>
#include "detcir.h"

#define iceil(X,Y)  (((X)+(Y)-1)/(Y))


void testimg(uint8_t *,int,int,char*);
#include <stdio.h>

static uint8_t *ialloc(int W,int H) { return malloc(W*H); }

static void greyscale
  (uint8_t *out, uint32_t *inp,int size, double cr, double cg, double cb)
{
  int R,G,B,Y;
  int i;
  for(i=0;i<size;i++)
  {
    R= (inp[i]>>16)&0xff;
    G= (inp[i]>>8)&0xff;
    B= (inp[i])&0xff;
    Y= R*cr + G*cg + B*cb;
    Y= Y<0 ? 0 : Y;
    Y= Y>255 ? 255: Y;
    out[i]= Y;
  }
}

static void threshold(uint8_t *d, int size, int V)
{
  int i;
  for(i=0;i<size;i++) if (d[i]<V) d[i]= 0;
}


static uint8_t* sobel(uint8_t *inp,int W,int H)
{
  int x,y,Px,Py;
  uint8_t *D;
  int V;
  
  D= ialloc(W,H);

  for(y=1;y<H-1;y++)
    for(x=1;x<W-1;x++)
    {
#define U(XX,YY) inp[(XX)+((YY)*W)]
      Py= - U(x-1,y-1) - 2*U(x,y-1) - U(x+1,y-1)
          + U(x-1,y+1) + 2*U(x,y+1) + U(x+1,y+1);

      Px=  -U(x-1,y-1)  +   U(x+1,y-1)
         -2*U(x-1,y)    + 2*U(x+1,y)
           -U(x-1,y+1)  +   U(x+1,y+1);
#undef U
      V= sqrt(Px*Px+Py*Py);
      if (V>255) V= 255;
      D[x+y*W]= V;
    }
  for(x=0;x<W;x++) { D[x+0*W]= D[x+1*W]; D[x+(H-1)*W]= D[x+(H-2)*W]; }
  for(y=1;y<H-1;y++) { D[0+y*W]= D[1+y*W]; D[(W-1)+y*W]= D[(W-2)+y*W]; }
  return D;
}

static void joinset(uint32_t *M, uint32_t a, uint32_t b)
{
  while(M[a]!=a) a= M[a];
  while(M[b]!=b) b= M[b];

  if (a<b) { int t= a; a= b; b= t; }
  M[a]= b;
}

static int scanfil(uint8_t *I, int W, int H, uint32_t **RS)
{
  uint32_t *M,*S,*N,R,A,B,C,L;
  M= malloc(sizeof(uint32_t)*(W*H/4+1));
  N= malloc(sizeof(uint32_t)*(W*H/4+1));
  S= calloc(W*H,sizeof(uint32_t));
  int i,x,y,K,lastset;
  M[0]= 0;
  lastset= 0;

#define input(x,y)  I[(x)+(y)*W]
#define output(x,y) S[(x)+(y)*W]

  for(y=0;y<H;y++)
   for(x=0;x<W;x++)
   {
     if (!input(x,y)) continue;
     A= B= C= L= 0;
     if (y!=0)
     {
       if (x!=0 && input(x-1,y-1)) A= output(x-1,y-1);
       if (input(x,y-1))  B= output(x, y-1);
       if (x!=W-1 && input(x+1,y-1)) C= output(x+1,y-1);
     }
     if (x!=0 && input(x-1,y)) L= output(x-1,y);

     if (A==0 && B==0 && C==0)
     {
       if (L==0)
       {
         K= ++lastset;
         output(x,y)= K;
         M[K]= K;
         continue;
       }
       output(x,y)= L;
       continue;
     }

     if (A==0) { if (B==0) R= C; else R= B; } else R= A;

     if (C!=0 && R!=C) { output(x+1,y-1)= R; joinset(M, C, R); }
     if (B!=0 && R!=B) { output(x, y-1)= R; joinset(M, B, R); }
     if (L!=0 && R!=L) { joinset(M, L, R); }
     output(x,y)= R;
   }
#undef input
#undef output

   for(i=1;i<=lastset;i++)
     while( M[M[i]]!=M[i] ) M[i]= M[M[i]];

   int la=1;
   N[0]= 0;
   for(i=1;i<=lastset;i++)
     if (M[i]==i) N[i]= la++;

   for(i=0;i<W*H;i++) S[i]= N[M[S[i]]];
   *RS= S;
   free(M); free(N);
   return la-1;
}


static double pdist(int Ax, int Ay,int Bx,int By)
{
  By-= Ay; Bx-= Ax;
  return sqrt(By*By+Bx*Bx);
}


static int get_maxdist_pair
   (uint8_t *S, int W, int H, int *rAX, int *rAY, int *rBX, int *rBY)
{
  int i,sz, Ax,Ay,Bx,By,Cx,Cy,Dx,Dy,t; 
  int dx,dy,dxS,dxE,dyS,dyE;
  double D,K,M;
  int fail;

  sz= W*H;
  for(i=0;i<sz;i++) if (S[i]) break;
  if (i==sz) return 1; 
  Ax= i%W; Ay= i/W;
           //  for(i++;i<sz;i++) if (S[i]) break;
  for(i=sz-1;i>=0;i--) if (S[i]) break;
  Bx= i%W; By= i/W;

  D= pdist(Ax,Ay,Bx,By);
  fail= 0;
  while(1)
  {
    M= 0;
    dyS= -1; dyE= 1; if (By==0) dyS=0; if (By==H-1) dyE=0;
    dxS= -1; dxE= 1; if (Bx==0) dxS=0; if (Bx==W-1) dxE=0;

    for(dy=dyS;dy<=dyE;dy++)
     for(dx=dxS;dx<=dxE;dx++)
    {
       if (dx==0 && dy==0) continue;
       Dx= Bx+dx; Dy= By+dy;
       if (!S[Dx+Dy*W]) continue; 
       K= pdist(Ax,Ay,Dx,Dy);
       if (K>M) { M=K; Cx= Dx; Cy= Dy;  }
    }

    if (M<=D)
    {
      fail++;
      if (fail==2) break;
      t= Ax; Ax= Bx; Bx= t;
      t= Ay; Ay= By; By= t;
    }
    else
    {
      D= M;
      Bx= Cx; By= Cy;
      fail= 0;
    }
  }

  *rAX= Ax; *rAY= Ay; *rBX= Bx; *rBY= By;

  return 0;
}

static int find_normal_intersect
  (uint8_t *S,int W,int H,int X0,int Y0,int X1,int Y1,int *rX2,int *rY2)
{
  int mX,mY,dX,dY; 
  int E, Eth, Esq, Edi; // error variable and constants for bresenham line
  int pX, pY; // current point and point update vector
  int adX, adY;
  int dXsq, dYsq, dXdi, dYdi; // square and diagonal updates
  int fX, fY; // found X, Y
  int found,count;  // found flag, count is used to travel in each direction
                    // once

  // these two define the slope of the normal, not the chord
  dY= -X1+ X0; dX=  Y1- Y0;
   // mid-point X and Y, this is the starting point
  mX= (X1+X0)/2; mY= (Y1+Y0)/2;

  adX= abs(dX); adY= abs(dY);
  if (adX>=adY)
  {
    Eth= adX-2*adY; Edi= -2*adX; Esq= 2*adY;
      // we update X on square moves, but not when error exceeds threshold
    dXsq= 1;  dXdi= 0; dYsq= 0;
    if (dX<0) { dX= -dX; dY= -dY; }
    if (dY<0) dYdi= -1; else dYdi= 1;
  }
  else
  {
    Eth= adY-2*adX; Edi= -2*adY; Esq= 2*adX;
    dYsq= 1; dYdi= 0; dXsq= 0;
    if (dY<0) { dY= -dY; dX= -dX; }
    if (dX<0) dXdi= -1; else dXdi= 1;
  }

  count= 0;
again:
  found= 0;
  if (count==2) return 1;

  pX= mX; pY= mY; E= 0;
  while(pX>=0 && pX<W && pY>=0 && pY<H)
  {
    if (S[pX+pY*W]) { fX= pX; fY= pY; found= 1; }
    if (E>Eth) { pX+= dXdi; pY+= dYdi; E+= Edi; }
    E+= Esq; pX+= dXsq; pY+= dYsq;
  }
  if (!found)
    { dXsq *= -1; dYsq *= -1; dXdi *= -1; dYdi *= -1; count++; goto again; }
  
  *rX2= fX; *rY2= fY;
  return 0;
}

static int circpix(uint8_t *S,int W,int H, int x,int y)
{
  if (x<0 || y<0 || x>=W || y>=H) return 0;
  return S[x+y*W];
}


static int test_circle(uint8_t *S,int W,int H, int cX,int cY, int r,int *Rcov)
{
  int count, V;
  int total, pixco;
  // from http://members.chello.at/~easyfilter/bresenham.html
  int x = -r, y = 0, err = 2-2*r;
  count= 0;
  total= 0;
  pixco= 0;
  do {
    count+= V= circpix(S,W,H,cX-x, cY+y); /*   I. Quadrant */
     if (V) pixco++;
    count+= V= circpix(S,W,H,cX-y, cY-x); /*  II. Quadrant */
     if (V) pixco++;
    count+= V= circpix(S,W,H,cX+x, cY-y); /* III. Quadrant */
     if (V) pixco++;
    count+= V= circpix(S,W,H,cX+y, cY+x); /*  IV. Quadrant */
     if (V) pixco++;
    total+= 4;
    r = err;
    if (r <= y) err += ++y*2+1;           /* e_xy+e_y < 0 */
    if (r > x || err > y) err += ++x*2+1; /* e_xy+e_x > 0 or no 2nd y-step */
  } while (x < 0);
  *Rcov= (pixco*100)/total;
  return count;
}

static int circle3points(int X0,int Y0,int X1,int Y1,int X2,int Y2,
                  int *rP, int *rQ, int *rR)
{
  double A0,A1,A2, B0,B1,B2, C0,C1,C2, D0,D1, E0,E1, K0,K1;
  double t, M, p, q, r;
  A0= 2*X0;   B0= 2*Y0;   C0= X0*X0+Y0*Y0;
  A1= 2*X1;   B1= 2*Y1;   C1= X1*X1+Y1*Y1;
  A2= 2*X2;   B2= 2*Y2;   C2= X2*X2+Y2*Y2;

  D0= 2*A0-A1-A2;  E0= 2*B0-B1-B2; K0= 2*C0-C1-C2;
  D1= 2*A1-A0-A2;  E1= 2*B1-B0-B2; K1= 2*C1-C0-C2;

  if (E0==0 || E1==0) return 1;
  
  t= E0/E1;
  M= K0 - t*K1;
  if (D0==D1*t) return 1;
  p= M / (D0-D1*t);
  q= (K0-D0*p)/E0;
  r= sqrt( (X0-p)*(X0-p)+(Y0-q)*(Y0-q) );

  *rP= p; *rQ= q; *rR= r;
  return 0;
}

static int is_circle_arc
   (uint8_t *S, int W, int H,
    int *Rcx,int *Rcy, int *Rr, int *Rcov, double *Rerr)
{
  int X0,X1,X2, Y0,Y1,Y2;
  int i,sz;
  int cX, cY, R;
  int N, T;
  int bv, bx, by, br, bcov, dx, dy, dr, x,y,r,v, cov;

  if (get_maxdist_pair(S,W,H, &X0, &Y0, &X1, &Y1)) return 0;
  if (find_normal_intersect(S,W,H,X0,Y0,X1,Y1,&X2,&Y2)) return 0;
  if (circle3points(X0,Y0,X1,Y1,X2,Y2, &cX, &cY,&R)) return 0;

  bv= 0;
  for(dy=-1;dy<2;dy++) for(dx=-1;dx<2;dx++) for(dr=-1;dr<2;dr++)
  {
     x= dx+cX; y= dy+cY; r= dr+R;
     v= test_circle(S,W,H, x,y,r,&cov);
     if (v>bv) { bv= v; bx= x; by= y; br= r; bcov= cov; }
  }
  if (bv==0) return 0;

  N= T= 0;
  sz= W*H;
  for(i=0;i<sz;i++)
    if (S[i]) { N++; T+= abs(pdist(i%W,i/W,bx,by)-r); }

  if (N==0) return 0;
 
  *Rcx= bx; *Rcy= by; *Rr= br; *Rcov= bcov; *Rerr= (double)T/N;
  return 1;
}

static void extract_part
  (uint8_t *shape, uint8_t *img, int index, uint32_t *parts, int W,int H)
{
  int i;
  int size;
  size= W*H;
  memset(shape, 0, size);
  for(i=0;i<size;i++) if (parts[i]==index) shape[i]= img[i];
}

static int bbox_area(uint8_t *S,int W,int H)
{
  int minX, minY, maxX, maxY,x,y;
  minX= W; minY= H; maxX= 0; maxY= 0;
  for(y=0;y<H;y++)
   for(x=0;x<W;x++)
    if (S[x+y*W])
    {
      minX= minX < x ? minX:x; minY= minY < y ? minY:y;
      maxX= maxX > x ? maxX:x; maxY= maxY > y ? maxY:y;
    }
  if (maxY<minY || maxX<minX) return 0;
  return (maxX-minX+1)*(maxY-minY+1);
}


static void makemask
  (uint8_t *S, int W,int H, detcirconf_t *CC,
   int *Rx, int *Ry, int *Rw, int *Rh)
{
  int x,y,bx,by;
  int sum,np;
  int bW, bH;
  int minX, minY,maxX, maxY;
  int extW, extH;
  int N, thr;

  N= CC->Mblk;
  thr= CC->Mbthr;

  bW= iceil(W,N); bH= iceil(H,N);

  minX= W; maxX= 0; minY= H; maxY= 0;

  for(by=0;by<bH;by++)
    for(bx=0;bx<bW;bx++)
    {
      np= 0;
      sum= 0;
      for(y=0;y<N && (y+by*N)<H;y++)
        for(x=0;x<N && (x+bx*N)<W;x++)
          { np++; sum += S[(x+bx*N)+(y+by*N)*W]; }
      if (np) sum/= np; 
      if (sum>=thr) 
      {
        if (minX>bx) minX= bx;
        if (maxX<bx) maxX= bx;
        if (minY>by) minY= by;
        if (maxY<by) maxY= by;
      }
    }

  minX= minX*N; maxX= maxX*N+N-1;
  minY= minY*N; maxY= maxY*N+N-1;

  extW= (maxX-minX+1)*CC->Mextend/100;
  extH= (maxY-minY+1)*CC->Mextend/100;

  minX -= extW; if (minX<0) minX= 0;
  maxX += extW; if (maxX>=W) maxX= W-1;
  minY -= extH; if (minY<0) minY= 0;
  maxY += extH; if (maxY>=H) maxY= H-1;

  *Rx= minX;
  *Ry= minY; 
  *Rw= maxX-minX+1;
  *Rh= maxY-minY+1;
}


static void make_input
   (uint32_t* cinp, int W,int H, detcirconf_t *CC,
    uint8_t **Rginp, int *Rw, int *Rh,int *Rofx, int *Rofy)
{
   uint8_t *ginp, *gc, *out;
   int Mx, My, Mw, Mh;
   int y;

   if (CC->mask_color)
   {
     gc= ialloc(W,H);
     greyscale(gc, cinp, W*H, CC->Mcr, CC->Mcg, CC->Mcb);
     threshold(gc, W*H, CC->Mcthr);
testimg(gc,W,H,"colthr");
     makemask(gc, W,H, CC, &Mx, &My, &Mw, &Mh);
     free(gc);
   }
   else
   {
     Mx= My= 0;
     Mw= W; Mh= H;
   }
   // FIXME: pad the greyscale image on four sides with 1 pixel
   // lines. this way I can get rid of the boundary condition
   // nonsense in the partitioning code.
   ginp= ialloc(Mw, Mh);
   cinp += Mx+My*W;
   out= ginp;
   for(y=0;y<Mh;y++)
   {
     greyscale(out, cinp, Mw, 0.3, 0.6, 0.1);
     out += Mw;
     cinp += W;
   }
   *Rginp= ginp;
   *Rw= Mw;
   *Rh= Mh;
   *Rofx= Mx;
   *Rofy= My;
}

static void color_transform(uint8_t *out, int size, uint32_t *inp)
{
  int i;
  int GG;
  int R, G, B; unsigned int P;
  for(i=0;i<size;i++)
  {
    P= inp[i];
    R= (P>>16)&0xff;
    G= (P>>8) & 0xff;
    B= (P) & 0xff;

    GG= R - G/2 - B/2;
    if (GG<0) GG= 0; 
    if (GG>255) GG= 255;
    out[i]= GG;
  }
}

static int gauss[25]=
 {
   1,  4,  7,  4,  1,
   4, 16, 26, 16,  4,
   7, 26, 41, 26,  7,
   4, 16, 26, 16,  4,
   1,  4,  7,  4,  1,
 };

static void gaussian_blur(uint8_t *out, uint8_t *inp, int W,int H)
{
  int N= 5;
  int C,g;
  int x,y,fx,fy,nx,ny,total,total_weight;
  
  C= N/2;
  for(y=0;y<H;y++)
   for(x=0;x<W;x++)
   {
     g= 0; total= 0; total_weight= 0;
     for(fy=-C;fy<=C;fy++) for(fx=-C;fx<=C;fx++,g++)
     {
       nx= fx+x; ny= fy+y;
       if (nx>=0 && nx<W && ny>=0 && ny<H)
       {
         total_weight+= gauss[g];
         total+= gauss[g]*inp[nx+ny*W];
       }
     }
     if (total_weight) out[x+y*W]= total/total_weight;
   }
}

int detect_circles
   (uint32_t *inp, int W, int H,detcirconf_t *CC,
    int *Rncircles, detcir_t **Rcircles)
{
  uint8_t *ginp; 
  uint8_t *sob;
  uint32_t *part; int nparts;
  detcir_t *circles, *C;
  int ncircles;
  int rv, i;
  uint8_t *shape;
  int cx,cy,r,cov; double er;
  uint8_t *blur;

  ncircles= 0; circles= NULL;
  rv= 0;

  ginp= ialloc(W,H);
  color_transform(ginp, W*H, inp);
  blur= ialloc(W,H);
  gaussian_blur(blur, ginp, W,H);

  testimg(ginp, W,H, "input");
  testimg(blur, W,H, "blur");

  sob= sobel(blur, W, H);
  testimg(sob, W,H, "sob");
  threshold(sob, W*H, CC->sobthr);
  testimg(sob, W,H, "sobthr");

  nparts= scanfil(sob, W,H, &part); 
  if (nparts==0) { printf("noparts\n"); rv= 1; goto errexit1; }

  circles= malloc(sizeof(*circles)*nparts);
  shape= ialloc(W,H);
  for(i=1;i<=nparts;i++)
  {
     extract_part(shape, sob, i, part, W,H);
     if (bbox_area(shape,W,H)<CC->minarea) continue;
     if (is_circle_arc(shape, W,H, &cx, &cy, &r, &cov, &er))
     {
       if (er>=CC->maxerr) continue;
       C= circles + ncircles;
       C->x= cx;
       C->y= cy;
       C->r= r;
       C->coverage= cov; 
       C->meanerr= er;
       ncircles++;
     }
  }

errexit1:
  free(ginp);
  free(sob);
  free(blur);
  if (rv)
  {
    if (circles) free(circles);
    *Rcircles= NULL;
    *Rncircles= 0;
  }
  else
  {
    *Rncircles= ncircles;
    *Rcircles= circles;
  }
  return rv;
}

void detcir_init(detcirconf_t *CC)
{
   CC->mask_color= 1;
   CC->Mcr= 1.0; CC->Mcg= -0.5; CC->Mcb= -0.5;
   CC->Mcthr= 180;
   CC->Mblk=16; CC->Mextend= 33; CC->Mbthr= 120;
   CC->sobthr= 200; CC->minarea= 25;
   CC->maxerr= 2.5;
   CC->gauss_rho= 1; CC->gauss_size= 5;
}
