#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include "rjpg.h"
#include "pngio.h"
#include "gfx.h"
#include "detcir.h"

typedef struct { int x,y,color; } mark_t;

typedef struct
{
  int type; // 'R'= RGB, 'g'= greyscale
  unsigned char *bdata;
  unsigned int  *idata;
  int w,h;
  mark_t *marks;
  int nmarks, smarks;

  int subtype;
  int iV[5];
} image_t;

void mark_point(image_t *img, int x,int y, int color)
{
  mark_t *M;
  if (img->nmarks==img->smarks)
  {
    img->smarks+= img->smarks/2 + 20;
    img->marks= realloc(img->marks, sizeof(img->marks[0])*img->smarks);
  }
  M= img->marks+img->nmarks++;
  M->x= x; M->y= y; M->color= color;
}


static image_t **images; int nimages, simages;
static image_t *cimage;

static image_t *bgimage;
static int bgfunc='X';

static image_t dpy;

image_t *image_new_dim(int type, int w, int h)
{
  image_t *img;
  img= calloc(1,sizeof(*img));
  img->type= type;
  img->w= w;
  img->h= h;
  if (type=='R') img->idata= calloc(img->w*img->h,sizeof(img->idata[0]));
            else img->bdata= calloc(img->w*img->h,sizeof(img->bdata[0]));
  return img;
}


image_t *image_new(int type, image_t *src)
{
  return image_new_dim(type, src->w, src->h);
}

int imax2(int a,int b) { return a>b?a:b; }
int imin2(int a,int b) { return a<b?a:b; }
int has_suffix(char *fn, char *sfx)
{
  int Lf, Ls;
  Lf= strlen(fn); Ls= strlen(sfx);
  if (Lf<=Ls) return 0;
  fn += Lf - Ls;
  return !strcasecmp(fn,sfx);
}


void get_canvas()
{
  unsigned int w,h,*B;
  gfx_canvas_get_buffer(Wcanvas, &w, &h, &B);
  dpy.w= w;
  dpy.h= h;
  dpy.idata= B;
}

void show_image(image_t *img)
{
  int cw, ch,ox,oy,iW,W,x,y,i,C;
  unsigned int *B;
  cw= imin2(img->w, dpy.w);
  ch= imin2(img->h, dpy.h);

  ox= (dpy.w-cw)/2;
  oy= (dpy.h-ch)/2;
  B= dpy.idata;
  W= dpy.w; iW= img->w;

  memset(dpy.idata, 0, dpy.w*dpy.h*sizeof(unsigned int));
  switch(img->type)
  {
  case 'R': for(y=0;y<ch;y++) for(x=0;x<cw;x++) 
               B[(y+oy)*W+(x+ox)]= img->idata[y*iW+x];
            break;
#define GRAY(X)  ( ((X)<<16) | ((X)<<8) | (X) )
  case 'g': 
    if (bgimage && bgfunc!='N')
    {
       for(y=0;y<ch;y++) for(x=0;x<cw;x++) 
       {
          unsigned int BP, DP;
          BP= bgimage->idata[y*bgimage->w+x];
          DP= GRAY(img->bdata[y*iW+x]);
          if (DP)
          {
            if (bgfunc=='X') BP ^= 0xFFffFF;
                        else BP = DP;
          }
          B[(y+oy)*W+(x+ox)]= BP;
       }
    } else 
    {
    for(y=0;y<ch;y++) for(x=0;x<cw;x++) 
               B[(y+oy)*W+(x+ox)]= GRAY(img->bdata[y*iW+x]);
    }
            break;
#undef GRAY
  }
  for(i=0;i<img->nmarks;i++)
  {
    static int colors[4]= { 0xFFFF0000, 0xFF00FF00, 0xFF0000FF, 0xFF00FFFF };
    x= img->marks[i].x; y= img->marks[i].y;
    if (x<0 || y<0) continue;
    if (x>=img->w || y>=img->h) continue;
    C= img->marks[i].color;
    B[(x+ox)+(y+oy)*W]= colors[C%4];
  }
  gfx_canvas_update(Wcanvas, 0,0, dpy.w, dpy.h); 
}


static void put_image_in_list(image_t *ni, image_t *oi, char *op, char *par)
{
  char buf[2000];
  int num, onum;
  if (nimages==simages)
  {
    simages += 50 + simages/2;
    images= realloc(images, sizeof(images[0])*simages);
  }
  if (oi)
  {
    int i;
    for(i=0;i<nimages;i++)
     if (images[i]==oi)  break;
    if (i==nimages) onum= -1;
           else onum= i;
  } else onum= -1;
  num= nimages;
  images[num]= ni;
  nimages++;
  if (onum!=-1) sprintf(buf,"[%d]= [%d] %s %s", num, onum, op, par);
          else  sprintf(buf,"[%d]= %s %s",num,op,par);
  gfx_insert(Wimages, -1, buf, NULL, NULL);
  cimage= ni;
  gfx_set_selected(Wimages, num);
  show_image(ni);
}


static void load_file(int color)
{
  char *fn,*e;
  image_t *img;
  static char *filetypes[]=
  { "Image Files",
    " JPEG Images (.jpg .jpeg)",
    " Portable Network Graphics (.png)", NULL };
  fn= gfx_select_file(0, NULL, filetypes);
  if (!fn) return ;
  img= calloc(1,sizeof(*img));
  if (has_suffix(fn,".png"))
  {
    if (color) e= read_png_image(fn, (unsigned int*) &img->w, 
                                     (unsigned int*) &img->h, &img->idata);
      else     e= read_png_graymap_noalpha(fn, (unsigned int*) &img->w, 
                                     (unsigned int*) &img->h, &img->bdata);
    if (e) { free(e); return ; }
  }
  else
  {
    color= 1;
    if (read_jpeg(fn,&img->w,&img->h,(int**) &img->idata)) return ;
  }
  if (color) img->type= 'R'; else img->type= 'g';
  char *z;
  while ((z=strchr(fn,'/'))) fn= z+1;
  put_image_in_list(img,NULL, "load ", fn);
}

static void save_file()
{
  char *fn,*e;
  static char *filetypes[]=
  { "Portable Network Graphics (.png)", NULL };
  if (!cimage) return ;
  fn= gfx_select_file(1, NULL, filetypes);
  if (!fn) return ;
  if (cimage->type=='g')
    e= write_png_g8(fn, cimage->w, cimage->h, cimage->bdata);
  else
  {
    int i, s;
    s= cimage->w*cimage->h;
    for(i=0;i<s;i++) cimage->idata[i] |= 0xFF000000;
    e= write_png_image(fn, cimage->w, cimage->h, cimage->idata);
  }
  if (e) fprintf(stderr,"write: %s\n",e);
}

static void save_combined()
{
  char *fn,*e;
  static char *filetypes[]=
  { "Portable Network Graphics (.png)", NULL };
  fn= gfx_select_file(1, NULL, filetypes);
  if (!fn) return ;
  {
    int i, s;
    s= dpy.w*dpy.h;
    for(i=0;i<s;i++) dpy.idata[i] |= 0xFF000000;
    e= write_png_image(fn, dpy.w, dpy.h, dpy.idata);
  }
  if (e) fprintf(stderr,"write: %s\n",e);
}

static void save_combined_big()
{
  char *fn,*e;
  unsigned int *dd,P;
  int x,y,ox,oy,cw,ch,dx,dy;
  static char *filetypes[]=
  { "Portable Network Graphics (.png)", NULL };
  if (!cimage) return ;
  fn= gfx_select_file(1, NULL, filetypes);
  if (!fn) return ;
 
#define BIG 4
  cw= imin2(cimage->w, dpy.w);
  ch= imin2(cimage->h, dpy.h);

  ox= (dpy.w-cw)/2;
  oy= (dpy.h-ch)/2;
  
  dd= malloc(sizeof(dd[0])*cw*BIG*ch*BIG);
  for(y=0;y<ch;y++)
   for(x=0;x<cw;x++)
   {
      P= dpy.idata[(x+ox)+(y+oy)*dpy.w];
      for(dy=0;dy<BIG;dy++) for(dx=0;dx<BIG;dx++)
        dd[ (BIG*x+dx) + (BIG*y+dy)*cw*BIG]= P | 0xFF000000;
   }
  e= write_png_image(fn, BIG*cw, BIG*ch, dd);
  if (e) fprintf(stderr,"write: %s\n",e);
  free(dd);
}

#include "gray.c"
#include "convo.c"
#include "gauss.c"
#include "center.c"
#include "threshold.c"
#include "sobel.c"
#include "mask.c"
#include "partition.c"
#include "circle.c"
#include "color.c"
#include "colortest.c"
#include "rmlone.c"

static void drapix(image_t *img,int x,int y)
{
  if (x<0 || y<0 || x>=img->w || y>=img->h) return ;
  img->idata[x+y*img->w] ^= 0xFFFFFF;
}

void dracir(image_t *img, int cx, int cy, int r)
{
  // from http://members.chello.at/~easyfilter/bresenham.html
  int x = -r, y = 0, err = 2-2*r;
  do {
    drapix(img, cx-x, cy+y); /*   I. Quadrant */
    drapix(img, cx-y, cy-x); /*  II. Quadrant */
    drapix(img, cx+x, cy-y); /* III. Quadrant */
    drapix(img, cx+y, cy+x); /*  IV. Quadrant */
    r = err;
    if (r <= y) err += ++y*2+1;           /* e_xy+e_y < 0 */
    if (r > x || err > y) err += ++x*2+1; /* e_xy+e_x > 0 or no 2nd y-step */
  } while (x < 0);
}

void cirdet()
{
  int i;
  detcirconf_t conf;
  image_t *img, *oi;
  detcir_t *C; int nC;
  if (!cimage || cimage->type!='R') return ;
  oi= cimage;
  detcir_init(&conf);
  
  if (detect_circles(cimage->idata,cimage->w,cimage->h,&conf,&nC, &C))
   return ;
  
  img= image_new('R', oi);
  memcpy(img->idata, oi->idata, sizeof(img->idata[0])*img->w*img->h);
  for(i=0;i<nC;i++)
  {
    if (C[i].r>=img->w/2) continue;
    if (C[i].coverage<5) continue;
    dracir(img, C[i].x, C[i].y, C[i].r);
    printf("CIRCLE %d,%d r=%d coverage=%d meanerr=%f\n",C[i].x,C[i].y,
           C[i].r, C[i].coverage, C[i].meanerr);
  }

  put_image_in_list(img, cimage, "circles", "");
}

void testimg(uint8_t *D,int W,int H,char *foo)
{
  image_t *img;
  img= image_new_dim('g', W,H);
  memcpy(img->bdata, D, W*H);
  put_image_in_list(img, cimage, "dbg", foo ? foo : "");
}


int main()
{
  gfx_event_t event;
  gfx_init(NULL,650,900, "Image Operations");
  gfx_SCREEN_0();

  get_canvas();

  gfx_canvas_set_zoom_params(Wcanvas,8,-1,0,dpy.w/3,dpy.h/3);
  int have_zoom= 0;
  
  while(1)
  {
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case GFXEV(Wcanvas,BUTTON_PRESS):
       if (event.button==3) 
       {
         if (have_zoom) { have_zoom= 0; gfx_canvas_hide_zoom(Wcanvas); 
                            gfx_canvas_end_report_mouse_position(Wcanvas); }
         else { have_zoom= 1; gfx_canvas_zoom(Wcanvas, event.x, event.y); 
                            gfx_canvas_begin_report_mouse_position(Wcanvas); }
       }
       break;
    case GFXEV(Wcanvas,POINTER_MOVE):
      if (have_zoom) { 
          gfx_canvas_zoom(Wcanvas, event.x, event.y);  }
       break;
                           
    case GFXEV_QUIT: goto done;
    case WloadRGB: load_file(1); break;
    case WloadGRAY: load_file(0); break;
    case Wsave: save_file(); break;
    case Wimages: cimage= images[event.key]; show_image(cimage); break;
    case Wgray:
       convert_gray(gfx_get_text(WgrRED), gfx_get_text(WgrGREEN),
                    gfx_get_text(WgrBLUE));
       break;
    case Wgauss:
      convolve_image(gaussian_filter(gfx_get_text(WgauN),
                                     gfx_get_text(WgauRHO)));
      break;
    case Wsobel: sobel(gfx_get_state(WsobNORM)); break;
    case Wsobeldir: sobel_dirs(); break;
    case Wthreshold: threshold(gfx_get_text(WthreVALUE)); break;
    case Wbg: 
       if (!cimage) break;
       if (cimage->type!='R') break;
       bgimage= cimage;
       break;
    case Wbgfunc:
       { char *f= "XSN"; bgfunc= f[(strchr(f,bgfunc)-f+1)%3]; }
       if (cimage!=bgimage && cimage && bgimage) show_image(cimage);
       break;
    case Wmask:
       genmask(gfx_get_text(WmaN), gfx_get_text(WmaTHR));
       break;
    case Wapplymask: applymask(); break;
    case WsaveCOM: save_combined(); break;
    case WsaveCOMbig: save_combined_big(); break;
    case Wpart: partition_image(); break;
    case Wcirc: find_circle(); break;
    case Wcolor: color_image(); break;
    case Wall: cirdet(); break;
    case Wcolortest: colortest(); break;
    case WrmLONE: rmlone(); break;
    }
  }
done:
  return 0;
}


