

#define DIST(x,y)  sqrt(x*x+y*y)

// This computes a triplet  Py,Br,Px
// where Py and Px are the gradients in the respective direction
// and Br is the intensity. Py and Px are signed 8 bit integers,
// Br is an unsigned 8 bit integer.
void sobel(int norm)
{
  int x,y,Px,Py,W,H;
  unsigned char *S;
  unsigned char *D;
  image_t *img;
  int Mv= 0;
  int dA,dB,dC;

  if (!cimage || cimage->type!='g') return ;
  img= image_new('g',cimage);
  W= img->w; H= img->h; D= img->bdata; S= cimage->bdata;

  if (norm)
  {

    for(y=1;y<H-1;y++) for(x=1;x<W-1;x++)
    {
      int A,B,C, a,b,c;
#define U(XX,YY) S[(XX)+((YY)*W)]
      a= U(x-1,y-1); b= U(x,y-1); c= U(x+1,y-1);
      A= U(x-1,y+1); B= U(x,y+1); C= U(x+1,y+1);

      Py= (A-a)+2*(B-b)+(C-c);

      a= U(x-1,y-1); A=U(x+1,y-1);
      b= U(x-1,y);   B=U(x+1,y);
      c= U(x-1,y+1); C=U(x+1,y+1);

      Px= (A-a)+2*(B-b)+(C-c);
#if 0
      Py= - U(x-1,y-1) - 2*U(x,y-1) - U(x+1,y-1)
          + U(x-1,y+1) + 2*U(x,y+1) + U(x+1,y+1);

      Px=  -U(x-1,y-1)  +   U(x+1,y-1)
         -2*U(x-1,y)    + 2*U(x+1,y)
           -U(x-1,y+1)  +   U(x+1,y+1);
#endif

#undef U

      if (Px<0) Px= -Px;
      if (Py<0) Py= -Py;
      Py= DIST(Px,Py);
      if (Py>Mv) Mv= Py;
    }
    if (Mv==0) Mv=255;
  }

  for(y=1;y<H-1;y++)
    for(x=1;x<W-1;x++)
    {
      int A,B,C, a,b,c;

#define U(XX,YY) S[(XX)+((YY)*W)]
      a= U(x-1,y-1); b= U(x,y-1); c= U(x+1,y-1);
      A= U(x-1,y+1); B= U(x,y+1); C= U(x+1,y+1);
      dA= A-a; dB= B-b; dC= C-c;

       Py= dA+2*dB+dC;
      // Py= dA*0.8+dB*2.2+dC*0.8;

      a= U(x-1,y-1); A=U(x+1,y-1);
      b= U(x-1,y);   B=U(x+1,y);
      c= U(x-1,y+1); C=U(x+1,y+1);
      dA= A-a; dB= B-b; dC= C-c;

       Px= dA+2*dB+dC;
      //Px= dA*0.8+dB*2.2+dC*0.8;
#undef U

      if (Px<0) Px= -Px;
      if (Py<0) Py= -Py;
      if (norm)
      {
        Py= DIST(Px,Py)*255/Mv;
      }
      else
      {
        Py= DIST(Px,Py);
        if (Py>255) Py= 255;
      }
      D[x+y*W]= Py;
    }
  for(x=0;x<W;x++)
  {
    D[x+0*W]= D[x+1*W];
    D[x+(H-1)*W]= D[x+(H-2)*W];
  }
  for(y=1;y<H-1;y++)
  {
    D[0+y*W]= D[1+y*W];
    D[(W-1)+y*W]= D[(W-2)+y*W];
  }
  put_image_in_list(img, cimage, "sobel", "");
}

void sobel_dirs()
{
  int W,H,i;
  unsigned int *S,*D;
  image_t *img;
  static unsigned int colors[]=
  {
#if 0
    0xFF0000, 0xff0000, 0xff0000, 0xff0000,
    0x0000ff, 0x0000ff, 0x0000ff, 0x0000ff,
#endif
    0x1f0000, 0x3e0000, 0x5d0000, 0x7c0000, 0x9b0000,
    0xba0000, 0xd90000, 0xff0000
  };
  if (!cimage || cimage->type!='R') return ;
  img= image_new('R',cimage);
  W= img->w; H= img->h; D= img->idata; S= cimage->idata;
  for(i=0;i<W*H;i++)
    if (S[i])  D[i]= colors[S[i]>>28]; 
         else  D[i]= 0;
  put_image_in_list(img, cimage, "sobel_dir", "");
}


#if 0
  classification stuff
      if (Px==0 && Py==0) { D[x+y*W]= 0; continue; }
      if (Py<0) { Py= -Py; Sy= -1; } else { Sy= 1; }
      if (Px<0) { Px= -Px; Sx= -1; } else { Sx= 1; }

#define QQ(A,B,C,D)  ((A<<24)|(B<<16)|(C<<8)|D)
      if (5*Py<2*Px) dir= QQ(2,2,6,6);       // closer to the X axis
      else if (5*Py>12*Px) dir= QQ(0,4,0,4); // closer to the Y axis
      else dir= QQ(1,3,7,5);                 // closer to the diagonal

      if (Sx==-1) dir &= 0xFFFF;  else dir >>= 16;
      if (Sy==-1) dir &= 0xFF;    else dir >>= 8;

      Py= Py+Px; if (Py>255) Py= 255;
      D[x+y*W]=  (dir<<28) | (Py<<8);
#undef QQ
#endif
